@@ -3,13 +3,17 @@ package Chemistry::OpenSMILES::Writer;
 use strict;
 use warnings;
 
-use Chemistry::OpenSMILES qw( is_aromatic is_chiral );
+use Chemistry::OpenSMILES qw(
+    is_aromatic
+    is_chiral
+    toggle_cistrans
+);
 use Chemistry::OpenSMILES::Parser;
 use Graph::Traversal::DFS;
-use List::Util qw(all uniq);
+use List::Util qw( any uniq );
 
 # ABSTRACT: OpenSMILES format writer
-our $VERSION = '0.7.0'; # VERSION
+our $VERSION = '0.8.5'; # VERSION
 
 require Exporter;
 our @ISA = qw( Exporter );
@@ -53,8 +57,10 @@ sub write_SMILES
             non_tree_edge => sub { my @sorted = sort { $vertex_symbols{$a} <=>
                                                        $vertex_symbols{$b} }
                                                      @_[0..1];
-                                   $rings->{$vertex_symbols{$sorted[0]}}
-                                           {$vertex_symbols{$sorted[1]}} =
+                                   $rings->{$vertex_symbols{$_[0]}}
+                                           {$vertex_symbols{$_[1]}} =
+                                   $rings->{$vertex_symbols{$_[1]}}
+                                           {$vertex_symbols{$_[0]}} =
                                         _depict_bond( @sorted, $graph ) },
 
             pre  => sub { my( $vertex, $dfs ) = @_;
@@ -92,41 +98,66 @@ sub write_SMILES
             next unless $atom->{chirality} =~ /^@@?$/;
 
             my @neighbours = $graph->neighbours($atom);
-            if( scalar @neighbours != 4 ) {
+            if( scalar @neighbours < 3 || scalar @neighbours > 4 ) {
                 # TODO: process also configurations other than tetrahedral
                 warn "chirality '$atom->{chirality}' observed for atom " .
                      'with ' . scalar @neighbours . ' neighbours, can only ' .
-                     'process tetrahedral chiral centers' . "\n";
+                     'process tetrahedral chiral centers with possible ' .
+                     'lone pairs' . "\n";
                 next;
             }
 
             my $chirality_now = $atom->{chirality};
             if( $atom->{chirality_neighbours} ) {
+                if( scalar @neighbours !=
+                    scalar @{$atom->{chirality_neighbours}} ) {
+                    warn 'number of neighbours does not match the length ' .
+                         "of 'chirality_neighbours' array, cannot process " .
+                         'such chiral centers' . "\n";
+                    next;
+                }
+
                 my %indices;
                 for (0..$#{$atom->{chirality_neighbours}}) {
-                    $indices{$vertex_symbols{$atom->{chirality_neighbours}[$_]}} = $_;
+                    my $pos = $_;
+                    if( scalar @{$atom->{chirality_neighbours}} == 3 && $_ != 0 ) {
+                        # Lone pair is always second in the chiral neighbours array
+                        $pos++;
+                    }
+                    $indices{$vertex_symbols{$atom->{chirality_neighbours}[$_]}} = $pos;
                 }
 
                 my @order_new;
                 # In newly established order, the atom from which this one
                 # is discovered (left hand side) will be the first, if any
                 if( $discovered_from{$atom} ) {
-                    push @order_new, $vertex_symbols{$discovered_from{$atom}};
+                    push @order_new,
+                         $indices{$vertex_symbols{$discovered_from{$atom}}};
                 }
                 # Second, there will be ring bonds as they are added the
                 # first of all the neighbours
                 if( $rings->{$vertex_symbols{$atom}} ) {
-                    push @order_new, sort { $a <=> $b }
+                    push @order_new, map  { $indices{$_} }
+                                     sort { $a <=> $b }
                                      keys %{$rings->{$vertex_symbols{$atom}}};
                 }
                 # Finally, all neighbours are added, uniq will remove duplicates
-                push @order_new, sort { $a <=> $b }
+                push @order_new, map  { $indices{$_} }
+                                 sort { $a <=> $b }
                                  map  { $vertex_symbols{$_} }
                                       @neighbours;
                 @order_new = uniq @order_new;
 
-                if( join( '', _permutation_order( map { $indices{$_} } @order_new ) ) ne
-                    '0123' ) {
+                if( scalar @order_new == 3 ) {
+                    # Accommodate the lone pair
+                    if( $discovered_from{$atom} ) {
+                        @order_new = ( $order_new[0], 1, @order_new[1..2] );
+                    } else {
+                        unshift @order_new, 1;
+                    }
+                }
+
+                if( join( '', _permutation_order( @order_new ) ) ne '0123' ) {
                     $chirality_now = $chirality_now eq '@' ? '@@' : '@';
                 }
             }
@@ -146,6 +177,7 @@ sub write_SMILES
         for my $i (0..$#symbols) {
             if( $rings->{$i} ) {
                 for my $j (sort { $a <=> $b } keys %{$rings->{$i}}) {
+                    next if $i > $j;
                     if( !@ring_ids ) {
                         # All 100 rings are open now. There is no other
                         # solution but to terminate the program.
@@ -237,7 +269,7 @@ sub _depict_bond
     my $bond = $graph->get_edge_attribute( $u, $v, 'bond' );
     return $bond if $bond ne '/' && $bond ne '\\';
     return $bond if $u->{number} < $v->{number};
-    return $bond eq '/' ? '\\' : '/';
+    return toggle_cistrans $bond;
 }
 
 # Reorder a permutation of elements 0, 1, 2 and 3 by taking an element
@@ -247,8 +279,13 @@ sub _depict_bond
 sub _permutation_order
 {
     # Safeguard against endless cycles due to undefined values
-    return unless scalar @_ == 4;
-    return unless all { defined } @_;
+    if( (scalar @_ != 4) ||
+        (any { !defined || !/^[0-3]$/ } @_) ||
+        (join( ',', sort @_ ) ne '0,1,2,3') ) {
+        warn '_permutation_order() accepts only permutations of numbers ' .
+             "'0', '1', '2' and '3', unexpected input received";
+        return 0..3; # Return original order
+    }
 
     while( $_[2] == 0 || $_[3] == 0 ) {
         @_ = ( $_[0], @_[2..3], $_[1] );
